{ *************************************************************************** }
{                                                                             }
{ Basic XPde screensaver                                                      }
{                                                                             }
{ Copyright (c) 2004 Sergey Mishin <mishinmail@mail.ru>                       }
{                                                                             }
{ This program is free software; you can redistribute it and/or               }
{ modify it under the terms of the GNU General Public                         }
{ License as published by the Free Software Foundation; either                }
{ version 2 of the License, or (at your option) any later version.            }
{                                                                             }
{ This program is distributed in the hope that it will be useful,             }
{ but WITHOUT ANY WARRANTY; without even the implied warranty of              }
{ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           }
{ General Public License for more details.                                    }
{                                                                             }
{ You should have received a copy of the GNU General Public License           }
{ along with this program; see the file COPYING.  If not, write to            }
{ the Free Software Foundation, Inc., 59 Temple Place - Suite 330,            }
{ Boston, MA 02111-1307, USA.                                                 }
{                                                                             }
{ *************************************************************************** }

unit uCanvas;

interface

uses
  SysUtils, Types, Classes, Variants, QTypes, QGraphics, QControls, QForms,
  QDialogs, QStdCtrls, QExtCtrls, Qt, libc, xlib;

type
 PProcInfoRec = ^TProcInfoRec;
 TProcInfoRec =
  record
   PID : Integer; { process id }
   Status : String; { process status }
   uName : String; { name of process initiator }
   CmdName : String; { name of process (no path) }
  end; { record }

type
  TfrmCanvas = class(TForm)
    Timer1: TTimer;
    imgLogo: TImage;
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure Timer1Timer(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  frmCanvas: TfrmCanvas;
  backbmp: TBitmap;
  x: Integer;
  y: Integer;
  xr: Integer;
  yr: Integer;
  xlast: integer;
  ylast: integer;
  progname: string;
  kssmode: boolean;

  
implementation

{$R *.xfm}

function EvtFilterFunc(XEvent: PXEvent): Boolean; cdecl;
begin
  case XEvent^.xtype of
      KeyPress:  frmCanvas.Close;
  end;
  Result := False;
end;

function ParseToSpace(var InStr : String) : String;
var
 OutStr : String;
begin
 OutStr := '';
 while (Length(Instr) > 0) and (InStr[1] = ' ') do
  Delete(Instr, 1, 1);
 while (Length(Instr) > 0) and (InStr[1] <> ' ') do
  begin
   OutStr := OutStr + InStr[1];
   Delete(InStr, 1, 1);
  end; { while }
 Result := OutStr;
end;

procedure TfrmCanvas.FormCreate(Sender: TObject);
var
   Root  : TWindow;
   Child : TWindow;
   Dummy : Integer;
   Background: TBitmap;
   OldX11EvtFilter: X11EventFilter;
begin
     kssmode:=false;
     if paramstr(1)='' then begin
     //KDE screensaver not defined, so will be XPde default screensaver
        kssmode:=False;
        imgLogo.Visible:=True;
     end
     else begin
        kssmode:=True;
        progname:=paramstr(1);
     end;

     OldX11EvtFilter := Application.SetX11EventFilter(EvtFilterFunc);

//Set initial position
     x:=0;
     y:=0;
     xr:=3;
     yr:=3;
     xlast := 0;
     ylast := 0;

//Set the canvas window parameters
     Top:=0;
     Left:=0;
     Height:=QForms.Screen.Height;
     Width:=QForms.Screen.Width;
     FormStyle:= fsStayOnTop;

//Set bitmap
     if not kssmode then begin
         backbmp :=TBitmap.Create;
         backbmp.Height:=Height;
         backbmp.Width:=Width;
     end;


//Draw desktop screenshot on the canvas window
     if kssmode then begin
           XGrabPointer(QtDisplay,  QWidget_winID(Self.Handle), 1, ButtonPressMask,
                       GrabModeAsync, GrabModeAsync, 0, 0, 0);
           if XQueryPointer(QtDisplay, QWidget_winID(Application.Desktop),
                          @Root, @Child, @Dummy, @Dummy, @Dummy, @Dummy, @Dummy)<>0 then begin
               Background := TBitmap.Create;
               QPixmap_grabWindow(Background.Handle, Root,0, 0, -1, -1);
               Bitmap:=Background;
               Background.Free;
           end;
     end;


end;

procedure TfrmCanvas.FormShow(Sender: TObject);
var
   s: string;
   winID: cardinal;
   result : integer;
begin
        
//Launch KDE screensaver in the specified xwindow
       if kssmode then begin
           winID:=QWidget_winID(ChildHandle);
           s:=progname+ ' --window-id '+IntToStr(winID)+ ' &';
            try
               libc.system(PChar(s));
            except

            end;
       end;
end;

procedure TfrmCanvas.FormClose(Sender: TObject; var Action: TCloseAction);
const
    PIPE_CMD : PChar = 'ps -eo pid,stat,user,args | grep ';
    PIPE_TYPE : PChar = 'r';

var
   f : PIOFile;
   PList : TList;
   PtrResult : Pointer;
   Found : Boolean;
   PSLine : String;
   CmdArr : array[0..512] of char;
   StrArr : array[0..1024] of char;
   ProcRec : PProcInfoRec;
   PID : Integer;
   Status : String;
   UName : String;
   CmdName : String;
begin
       if kssmode then begin
            PList := TList.Create;
            StrPCopy(CmdArr, PIPE_CMD);
            StrPCopy(StrArr, progname);
            StrCat(CmdArr, StrArr);
            //Build list of working screensavers
            f := popen(CmdArr, PIPE_TYPE);
            if F <> nil then begin
                 repeat
                  PtrResult := fgets(StrArr, 1024, F);
                     if PtrResult <> nil then begin
                         PSLine := StrPas(StrArr);
                         PSLine := Copy(PSLine, 1, Length(PSLine) - 1);
                         PID := StrToInt(ParseToSpace(PSLine));
                         Status := ParseToSpace(PSLine);
                         UName := ParseToSpace(PSLine);
                         CmdName := ParseToSpace(PSLine);
                         CmdName := ExtractFileName(CmdName);
                         Found := CmdName = progname;
                         if Found then begin
                              ProcRec := New(PProcInfoRec);
                              ProcRec.PID := PID;
                              ProcRec.Status := Status;
                              ProcRec.UName := UName;
                              ProcRec.CmdName := CmdName;
                              PList.Add(ProcRec);
                         end;
                     end;
                   until PtrResult = nil;
             end;
             pclose(F);
             ProcRec := PList.Items[PList.Count-1];  //Our screensaver must have
             try
                with ProcRec^ do kill(PID, SIGTERM);
             except

             end;
             PList.Free;
       end;
       Action:=caFree;
end;

procedure TfrmCanvas.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var  xnow : Integer;
     ynow : Integer;
begin
// Get current position
    xnow := x;
    ynow := y;

// On first move, simply record position
    if (xlast = 0) and (ylast = 0) then begin
        xlast:=xnow;
        ylast:=ynow;
        exit;
    end;
// Quit only if mouse actually changes position
      if (xnow <> xlast) or (ynow <> ylast) then close;

end;

procedure TfrmCanvas.Timer1Timer(Sender: TObject);
begin
     XGrabKeyboard(QtDisplay, QWidget_winID(Self.Handle), Ord(True),
                   GrabModeAsync, GrabModeAsync, CurrentTime);
     if not kssmode then begin
         x:=x+xr;
         y:=y+yr;
         if x < 0 then xr:=3;
         if x > Width-imgLogo.Picture.Bitmap.Width then xr:=-3;
         if y < 0 then yr:=3;
         if y > Height-imgLogo.Picture.Bitmap.Height then yr:=-3;
         FormPaint(nil);
     end;
end;

procedure TfrmCanvas.FormPaint(Sender: TObject);
begin

    if not kssmode then begin
        backbmp.Canvas.Brush.Color:=clBlack;
        backbmp.Canvas.FillRect(ClientRect);
        backbmp.Canvas.Draw(x,y,imgLogo.Picture.Bitmap);
        Canvas.Draw(0,0,backbmp);
    end;
   
end;


procedure TfrmCanvas.FormClick(Sender: TObject);
begin
     close;
end;

end.
